/*
   Pipe redirector.

   This module will try to connect to one of listed named pipes and, optionaly,
   redirect to it's one (server) named pipes.
*/

#include <stdio.h>
#include <process.h>
#include <string.h>
#include <ctype.h>
#define INCL_DOSNMPIPES
#define INCL_DOSSEMAPHORES
#define INCL_DOSERRORS
#define INCL_DOSMISC
#define INCL_DOSPROCESS
#include <os2.h>
#include "piper.h"
#include "debug.h"     // Must be the last.

#define _WRITE_BUF_SIZE          1024
#define _READ_BUF_SIZE           1024

struct _PIPER {
  HPIPE                hPipe;
  PSZ                  pcPipes;
  ULONG                ulLostPipeTime;
  ULONG                ulReconnectPeriod;
  PSZ                  pszServerPipes;
  HEV                  hevServerPipes;
  ULONG                cServerPipes;
  PPRFNUSER            fnUser;
  ULONG                ulWriteBufSize;
  ULONG                ulReadBufSize;

  PHPIPE               phServerPipes;
  PCHAR                pcReadBuf;
  ULONG                ulReadBufPos;
  PSZ                  pszCurPipe;
                       // Currently connected pipe name from the list pcPipes.
};

typedef struct _PIPER PIPER;


static ULONG _expandPipeName(ULONG cbBuf, PCHAR pcBuf, PSZ pszName)
{
  int        cBytes;

  while( isspace( *pszName ) )
    pszName++;

  if ( memicmp( pszName, "\\PIPE\\", 6 ) == 0 )
    pszName += 6;

  cBytes = _snprintf( pcBuf, cbBuf, "\\PIPE\\%s", pszName );
  if ( cBytes < 0 )
    return 0;

//  strupr( &pcBuf[6] );

  return cBytes;
}

static VOID _destroyServerPipes(PPIPER pPiper)
{
  ULONG      ulIdx;

  if ( pPiper->phServerPipes == NULL )
  {
    debugCP( "Not created" );
    return;
  }

  for( ulIdx = 0; ulIdx < pPiper->cServerPipes; ulIdx++ )
    if ( pPiper->phServerPipes[ulIdx] != NULLHANDLE )
    {
      DosClose( pPiper->phServerPipes[ulIdx] );
      pPiper->phServerPipes[ulIdx] = NULLHANDLE;
    }

  free( pPiper->phServerPipes );
  pPiper->phServerPipes = NULL;
}

static BOOL _createServerPipes(PPIPER pPiper)
{
  ULONG      ulRC = NO_ERROR;
  ULONG      ulIdx;

  if ( pPiper->cServerPipes == 0 )
    return TRUE;

  if ( pPiper->phServerPipes != NULL )
  {
    debugCP( "Already created" );
    return TRUE;
  }

  pPiper->phServerPipes = calloc( pPiper->cServerPipes, sizeof(HPIPE) );
  if ( pPiper->phServerPipes == NULL )
  {
    debugCP( "Not enough memory" );
    return FALSE;
  }

  for( ulIdx = 0; ulIdx < pPiper->cServerPipes; ulIdx++ )
  {
    ulRC = DosCreateNPipe( pPiper->pszServerPipes, &pPiper->phServerPipes[ulIdx],
                           NP_NOINHERIT | NP_ACCESS_OUTBOUND,
                           NP_NOWAIT | NP_TYPE_BYTE | pPiper->cServerPipes,
                           pPiper->ulWriteBufSize, 0, 0 );
    if ( ulRC != NO_ERROR )  
    {
      debug( "DosCreateNPipe(\"%s\",...), rc = %u", pPiper->pszServerPipes, ulRC );
      break;
    }

    ulRC = DosSetNPipeSem( pPiper->phServerPipes[ulIdx],
                           (HSEM)pPiper->hevServerPipes, ulIdx );
    if ( ulRC != NO_ERROR )
    {
      debug( "DosSetNPipeSem(), rc = %u", ulRC );
      break;
    }

    ulRC = DosConnectNPipe( pPiper->phServerPipes[ulIdx] );
    if ( ulRC == ERROR_PIPE_NOT_CONNECTED )
      ulRC = NO_ERROR;
    else if ( ulRC != NO_ERROR )
    {
      debug( "DosConnectNPipe(), rc = %u", ulRC );
      break;
    }
  }

  if ( ulRC != NO_ERROR )
  {
    _destroyServerPipes( pPiper );

    if ( pPiper->fnUser != NULL )
      pPiper->fnUser( pPiper, PREVENT_PIPECREATEERROR, pPiper->pszServerPipes );

    return FALSE;
  }

  debug( "%lu server pipe(s) %s created",
         pPiper->cServerPipes, pPiper->pszServerPipes );
  return TRUE;
}



ULONG prInit(PPIPER *ppPiper, PPRINIT pInit)
{
  CHAR       acName[_MAX_PATH];
  ULONG      ulRC;
  ULONG      cBytes;
  ULONG      ulRes = PRRC_OK;
  PPIPER     pPiper = calloc( 1, sizeof(PIPER) );
  
  if ( pPiper == NULL )
    return PRRC_NOTENOUGHMENORY;

  pPiper->pcReadBuf = malloc( pInit->ulReadBufSize );

  if ( pPiper->pcReadBuf == NULL )
    ulRes = PRRC_NOTENOUGHMENORY;
  else
  do
  {
    // Store server pipes name and create server pipes event semaphore.

    if ( ( pInit->pszServerPipe == NULL ) || ( *pInit->pszServerPipe == '\0' )
         || ( pInit->cServerPipes == 0 ) )
      pPiper->cServerPipes = 0;
    else
    {
      cBytes = _expandPipeName( sizeof(acName), acName, pInit->pszServerPipe );
      if ( cBytes == 0 )
      {
        ulRes = PRRC_INVALIDSRVPIPENAME;
        break;
      }

      pPiper->pszServerPipes = strdup( acName );
      if ( pPiper->pszServerPipes == NULL )
      {
        ulRes = PRRC_NOTENOUGHMENORY;
        break;
      }
      debug( "Server pipes name: %s", pPiper->pszServerPipes );

      ulRC = DosCreateEventSem( NULL, &pPiper->hevServerPipes,
                                DC_SEM_SHARED | DCE_AUTORESET, FALSE );
      if ( ulRC != NO_ERROR )
      {
        debug( "DosCreateEventSem(), rc = %u", ulRC );
        ulRes = PRRC_SRVPIPECREATEERR;
        break;
      }

      pPiper->cServerPipes = pInit->cServerPipes;
    }  // if ( ( pszServerPipe != NULL ) && ( cServerPipes != 0 ) )


    // Store the list of pipes to which we will connect

    if ( pInit->pcPipes != NULL )
    {
      PSZ      pszScan = pInit->pcPipes;
      PSZ      pszNew;
      ULONG    cbPipes = 0;

      while( *pszScan != '\0' )
      {
        // Expand pipe name to \PIPE\xxxxx
        cBytes = _expandPipeName( sizeof(acName), acName, pszScan );
        if ( cBytes == 0 )
        {
          ulRes = PRRC_INVALIDPIPENAME;
          break;
        }

        if ( ( pPiper->pszServerPipes != NULL ) &&
             ( stricmp( acName, pPiper->pszServerPipes ) == 0 ) )
        {
          ulRes = PRRC_NAMECOLLISION;
          break;
        }

        // Add name to the list
        cBytes++;      // '\0' terminator for name.
        pszNew = realloc( pPiper->pcPipes,
                          cbPipes + cBytes + 1 ); // +1 : second '\0' for list.
        if ( pszNew == NULL )
        {
          debugCP( "Not enough memory" );
          ulRes = PRRC_NOTENOUGHMENORY;
          break;
        }
        pPiper->pcPipes = pszNew;

        debug( "Store Weasel log pipe name (%lu): %s", cBytes, acName );
        memcpy( &pszNew[cbPipes], acName, cBytes );
        cbPipes += cBytes;
        pszNew[cbPipes] = '\0';

        // Jump to next name
        pszScan = strchr( pszScan, '\0' ) + 1;
      }  // while( *pszScan != '\0' )

    }  // if ( pcPipes != NULL )
  }
  while( FALSE );

  if ( ulRes != PRRC_OK )
  {
    prDone( pPiper );
    *ppPiper = NULL;
  }
  else
  {
    pPiper->ulReconnectPeriod = pInit->ulReconnectPeriod;
    pPiper->fnUser = pInit->fnUser;
    pPiper->ulReadBufSize = pInit->ulReadBufSize;
    pPiper->ulWriteBufSize = pInit->ulWriteBufSize;

    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &pPiper->ulLostPipeTime,
                     sizeof(ULONG) );
    pPiper->ulLostPipeTime -= pPiper->ulReconnectPeriod;
    *ppPiper = pPiper;
  }

  return ulRes;
}

VOID prDone(PPIPER pPiper)
{
  if ( pPiper == NULL )
    return;

  _destroyServerPipes( pPiper );

  if ( pPiper->hevServerPipes != NULLHANDLE )
    DosCloseEventSem( pPiper->hevServerPipes );

  if ( pPiper->pszServerPipes != NULL )
    free( pPiper->pszServerPipes );

  if ( pPiper->pcPipes != NULL )
    free( pPiper->pcPipes );

  if ( pPiper->hPipe != NULLHANDLE )
    DosClose( pPiper->hPipe );

  if ( pPiper->pcReadBuf != NULL )
    free( pPiper->pcReadBuf );

  free( pPiper );
}

VOID prProcess(PPIPER pPiper)
{
  ULONG      ulRC;
  ULONG      ulActual, cbBuf;
  ULONG      ulIdx;
  HPIPE      hNewPipe;
  PCHAR      pcReadBuf;

  if ( pPiper == NULL )
    return;

  if ( ( pPiper->hPipe == NULLHANDLE ) && ( pPiper->pcPipes != NULL ) )
  {
    // Client pipe not connected

    ULONG    ulTime;
    PSZ      pszCurPipe = (PSZ)pPiper->pcPipes;

    if ( pPiper->hPipe == NULLHANDLE )
    {
      DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulTime, sizeof(ULONG) );
      if ( ( ulTime - pPiper->ulLostPipeTime ) < pPiper->ulReconnectPeriod )
        return;
    }

    while( *pszCurPipe != '\0' )
    {
      // Try to open pipe...
      ulRC = DosOpen( pszCurPipe, &hNewPipe, &ulActual, 0, FILE_NORMAL,
                      OPEN_ACTION_FAIL_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
                      OPEN_FLAGS_FAIL_ON_ERROR | OPEN_FLAGS_SEQUENTIAL |
                      OPEN_FLAGS_NOINHERIT | OPEN_SHARE_DENYNONE |
                      OPEN_ACCESS_READONLY, NULL );
      if ( ulRC == NO_ERROR )
      {
        debug( "Connected to %s", pszCurPipe );
        pPiper->pszCurPipe = pszCurPipe;
        pPiper->hPipe = hNewPipe;

        if ( pPiper->fnUser != NULL )
          pPiper->fnUser( pPiper, PREVENT_CONNECTED, pszCurPipe );

        ulRC = DosSetNPHState( hNewPipe, NP_NOWAIT );
        if ( ulRC != NO_ERROR )
          debug( "DosSetNPHState(), rc = %lu", ulRC );

        _createServerPipes( pPiper );

        break;
      }
      debug( "Could not connect to %s", pszCurPipe );

      // Jump to the next name
      pszCurPipe = strchr( pszCurPipe, '\0' ) + 1;
    }

    if ( pPiper->hPipe == NULLHANDLE )
    {
      pPiper->ulLostPipeTime = ulTime;
      if ( pPiper->fnUser != NULL )
        pPiper->fnUser( pPiper, PREVENT_CONNECTERROR, NULL );
      return;
    }
  }


  // Check (close) events on server pipes

  if ( ( pPiper->hevServerPipes != NULLHANDLE ) &&
       ( pPiper->cServerPipes != 0 ) )
  {
    ulRC = DosWaitEventSem( pPiper->hevServerPipes, SEM_IMMEDIATE_RETURN );
    if ( ulRC != NO_ERROR )
    {
      if ( ulRC != ERROR_TIMEOUT )
        debug( "DosWaitEventSem(), rc = %u", ulRC );
    }
    else
    {
      PPIPESEMSTATE      paPipeState, pPipeState;
      ULONG              cbPipeState = (pPiper->cServerPipes + 2) * sizeof(PIPESEMSTATE);

      paPipeState = calloc( 1, cbPipeState );
      if ( paPipeState == NULL )
      {
        debugCP( "Not enough memory" );
        return;
      }

      // Query information about pipes that are attached to the semaphore. 
      ulRC = DosQueryNPipeSemState( (HSEM)pPiper->hevServerPipes, paPipeState,
                                    cbPipeState );
      if ( ulRC != NO_ERROR )
        debug( "DosQueryNPipeSemState(), rc = %u", ulRC );
      else
      {
        for( ulIdx = 0, pPipeState = paPipeState; ulIdx < pPiper->cServerPipes;
             ulIdx++, pPipeState++ )
        {
          switch( pPipeState->fStatus )
          {
  /*          case NPSS_WSPACE:
                break;*/

            case NPSS_CLOSE:
              DosDisConnectNPipe( pPiper->phServerPipes[pPipeState->usKey] );
              DosConnectNPipe( pPiper->phServerPipes[pPipeState->usKey] );
              break;
          }
        }
      }

      free( paPipeState );
    }
  }  // if ( hevServerPipes != NULLHANDLE )


  // Read data from the pipe

  pcReadBuf = pPiper->pcReadBuf;
  ulRC = DosRead( pPiper->hPipe, &pcReadBuf[pPiper->ulReadBufPos],
                  pPiper->ulReadBufSize - pPiper->ulReadBufPos, &cbBuf );
  if ( ulRC == ERROR_NO_DATA )
    return;

  if ( ( ulRC != NO_ERROR ) || ( cbBuf == 0 ) )
  {
    // Client pipe is disconnected

    if ( ulRC != NO_ERROR )
      debug( "DosRead(), rc = %lu", ulRC );

    debugCP( "Close client pipe" );
    if ( pPiper->fnUser != NULL )
      pPiper->fnUser( pPiper, PREVENT_DISCONNECTED, pPiper->pszCurPipe );
    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &pPiper->ulLostPipeTime,
                     sizeof(ULONG) );
    DosClose( pPiper->hPipe );
    pPiper->hPipe = NULLHANDLE;
    pPiper->pszCurPipe = NULL;
    _destroyServerPipes( pPiper );
    return;
  }


  // Write data to the output (server) pipes

  for( ulIdx = 0; ulIdx < pPiper->cServerPipes; ulIdx++ )
    DosWrite( pPiper->phServerPipes[ulIdx], &pcReadBuf[pPiper->ulReadBufPos],
              cbBuf, &ulActual );


  // Split input data on lines and call user function
  {
    PCHAR    pcEOL;
    ULONG    ulReadBufPos = pPiper->ulReadBufPos + cbBuf;

    while( TRUE )
    {
      pcEOL = memchr( pcReadBuf, '\n', ulReadBufPos );
      if ( pcEOL != NULL )
      {
        if ( ( pcEOL > pcReadBuf ) && ( *(pcEOL - 1) == '\r' ) )
          *(pcEOL - 1) = '\0';
        else
          *pcEOL = '\0';

        if ( pPiper->fnUser != NULL )
          pPiper->fnUser( pPiper, PREVENT_INPUTLINE, pcReadBuf );

        pcEOL++;
        ulReadBufPos -= ( pcEOL - pcReadBuf );
        pcReadBuf = pcEOL;
      }
      else
      {
        if ( pcReadBuf != pPiper->pcReadBuf )
          // Move the tail without a line break at the beginning of the buffer.
          memcpy( pPiper->pcReadBuf, pcReadBuf, ulReadBufPos );
        else if ( ulReadBufPos == pPiper->ulReadBufSize )
          // Received line is too long.
          ulReadBufPos = 0;

        break;
      }
    }

    pPiper->ulReadBufPos = ulReadBufPos;
  }
}

BOOL prIsConnected(PPIPER pPiper)
{
  return ( pPiper != NULL ) && ( pPiper->hPipe != NULLHANDLE );
}
